# Copyright European Space Agency, 2013

from __future__ import print_function

import os
from six.moves import map
import unittest
from numpy.testing.utils import assert_equal
from nose.plugins.attrib import attr

from auromat.mapping.spacecraft import getMapping
from auromat.resample import resample
from auromat.mapping import miracle
from auromat.mapping.mapping import BoundingBox, GenericMapping
from auromat.utils import outline
from auromat.util import coroutine
import numpy as np

try:
    from auromat.draw import drawMLatMLTPolar, drawStereographic,\
        drawStereographicMLatMLT, drawHeatmap, drawHeatmaps,\
        drawHorizon, drawIndxPlot, saveFig, drawConstellations,\
        drawParallelsAndMeridians, drawScanLinesCo,\
        drawScanLinesMLatMLTCo
    from mpl_toolkits.basemap import Basemap
    import matplotlib.pyplot as plt
    from auromat.draw_helpers import loadFigImage
except ImportError as e:
    print(repr(e))
    # import is optional so that the test module import doesn't fail if
    # the tests are not actually run and matplotlib is not installed

@attr('slow')
class Test(unittest.TestCase):
    def _testHorizonImage(self):
        m = _getMappingNorth()
        saveFig('horizon.jpg', drawHorizon(m))
        
    def _testColors(self):
        m = _getMappingNorth()
        m = resample(m, arcsecPerPx=100, method='mean')
        saveFig('test_white.png', drawStereographic(m, bottomTitle='foo'), bgcolor='white')
        saveFig('test_black.png', drawStereographic(m, bottomTitle='foo'), bgcolor='black')
        saveFig('test_polar_white.png', drawMLatMLTPolar(m), bgcolor='white')
        saveFig('test_polar_black.png', drawMLatMLTPolar(m), bgcolor='black')
               
    def testIndxPlot(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        figax = drawIndxPlot(imagePath, wcsPath=wcsPath, useWebCatalog=True, webCatalogLimit=200)
        saveFig('test_indx.png', figax)
        
    def _testConstellationPlot(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        
        m = getMapping(imagePath, wcsPath)
        figax = loadFigImage(imagePath)
        drawConstellations(figax, wcsPath, clipPoly=outline(~m.intersectsEarth))
        saveFig('test_constellations.jpg', figax)
    
    def testParallelsMeridiansPlot(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        m = getMapping(imagePath, wcsPath, altitude=0, fastCenterCalculation=True)
        figax = loadFigImage(imagePath)
        drawParallelsAndMeridians(m, figax=figax)
        drawConstellations(figax, wcsPath, clipPoly=outline(~m.intersectsEarth))
        saveFig('test_parallelsmeridians.jpg', figax)
    
    def _testParallelsMeridiansPlotOptimized(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        # As we precalculated the bounding box and the lat/lon label position
        # we only need to access the 'latCenter', 'lonCenter' and 'intersectsEarth'
        # attributes of the spacecraft mapping. This means that we don't need
        # any sanitization applied on the corner coordinate arrays ('lats','lons').
        # For this reason, and because we don't use masking, we set nosanitize=True
        # and can cut the required time in half. 
        m = getMapping(imagePath, wcsPath, altitude=0, fastCenterCalculation=False, nosanitize=True)
        
        # the following values must be precalculated to use the optimization
        # the bounding box at altitude=0 without any masking
        bb = BoundingBox(latSouth=49.3401757697, lonWest=-116.368770925, latNorth=64.288454984, lonEast=-91.8890098192)
        # center of bounding box at altitude=0 when masked below 10deg elevation
        labelLat, labelLon = 53.133, -98.684
        
        figax = loadFigImage(imagePath)
        drawParallelsAndMeridians(figax, m, boundingBox=bb, labelLat=labelLat, labelLon=labelLon)
        drawConstellations(figax, wcsPath, clipPoly=outline(~m.intersectsEarth))
        saveFig('test_parallelsmeridians_optimized.jpg', figax)
    
    def _testBBPole2(self):
        bb = BoundingBox(latSouth=35.3446724767, lonWest=-180.0, latNorth=90.0, lonEast=180.0)
        m = _getMappingNorth()
        m = resample(m, arcsecPerPx=100, method='mean')
        saveFig('test_bb.png', drawStereographic(m, boundingBox=bb))
    
    def _testStereographicMap(self):
        m = _getMappingSouth()
        m = m.maskedByElevation(10)
        m = resample(m, arcsecPerPx=100, method='mean')
        saveFig('test_stereo.svg', drawStereographic(m), dpi=200)
    
    def _testMLatMLTPolarMapNorthPole(self):
        m = _getMappingNorth()
        m.checkGuarantees()
        #m = m.maskedByElevation(10)
        #m.checkGuarantees()
        m = resample(m, arcsecPerPx=100, method='mean')
        m.checkGuarantees()
        saveFig('test_mlatmlt_polar_north.png', drawMLatMLTPolar(m))
         
    def _testMLatMLTPolarMapBothHemispheres(self):
        m = _getMappingSouth()
        m.checkGuarantees()
        #m = m.maskedByElevation(10)
        #m.checkGuarantees()
        m = resample(m, arcsecPerPx=100, method='mean')
        m.checkGuarantees()
        seqbb = BoundingBox(latSouth=-61.150846231, lonWest=142.622725698, latNorth=6.84984918353, lonEast=-116.820615123)
        saveFig('test_mlatmlt_polar_seqbb.png', drawMLatMLTPolar(m, boundingBox=seqbb))
               
    def _testScanlineMap(self):
        imagePath = getResourcePath('ISS029-E-8492.jpg')
        import gc, psutil
        proc = psutil.Process()
        
        from auromat.mapping.mapping import BaseMapping 
        def _getmapping(frame):
            gc.collect()
            if frame > 8494:
                # at this place we only have exactly 1 resampled mapping floating around
                assert_equal(len(list(filter(lambda o: isinstance(o, BaseMapping), gc.get_objects()))), 1)                
                
            print('Process memory:', proc.get_memory_info()[0]/1024/1024, 'MB')
            wcsPath = getResourcePath('seq/ISS029-E-%i.wcs' % frame)
            mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True, identifier=str(frame))
            mapping = mapping.maskedByElevation(10)
            # save some memory by removing additional data
            mapping = GenericMapping.fromMapping(mapping)
            print('SENDING MAPPING', frame, 'TO COROUTINES')            
            return mapping
        
        co = drawScanLinesCo('test_scanlines.png', arcsecPerPx=100, bgcolor='white')
        mappings = map(_getmapping, range(8493,8503))
        coroutine.broadcast(mappings, co)
                
    def _testScanlineMapBug(self):       
        """
        Not handling degenerate contours led to exceptions in auromat.utils.polygonArea
        as the polygon had just 0 or 1 vertices.
        See also the comment in auromat.utils.outline_skimage. 
        """
        imagePath = getResourcePath('ISS029-E-8492.jpg')
        def seq():
            for frame in range(229356,229360):
                wcsPath = getResourcePath('seq2/ISS030-E-%i.wcs' % frame)
                mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
                mapping = mapping.maskedByElevation(10)
                print('SENDING MAPPING', frame, 'TO COROUTINES')
                yield mapping
        
        co = drawScanLinesMLatMLTCo('test_scanlines2.png', arcsecPerPx=100, lineWidthFactor=3, bgcolor='white')
        coroutine.broadcast(seq(), co)
        
    def _testScanlineMapBug2(self):       
        """
        polygon masking was wrong, didn't honour existing mask
        note that this test has to be manually run and visually inspected
        """
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        def seq():
            for frame in range(102170,102172):
                wcsPath = getResourcePath('seq3/ISS030-E-%i.wcs' % frame)
                mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
                mapping = mapping.maskedByElevation(10)
                print('SENDING MAPPING', frame, 'TO COROUTINES')
                yield mapping
        
        co = drawScanLinesCo('test_scanlines3.png', arcsecPerPx=100, lineWidthFactor=3, bgcolor='white')
        coroutine.broadcast(seq(), co) 
               
    def _testHoleBug(self):
        """
        Triggers a (now fixed) bug that occured when 'mean' resampling led to
        single-polygon holes where surrounding polygons were well-defined.
        A hole is a polygon such that all corners are defined and just the polygon
        center data was missing (color etc.). The issue was that those holes were
        not filtered and led to subsequent errors. The fix was to filter not by whether
        all corners are defined but whether the color is defined as this also guarantees
        that the corners are defined.        
        """
        m = _getMappingNorth()
        m = resample(m, arcsecPerPx=200, method='mean')
        # saving as svg without rasterization will lead to errors when trying to use
        # polygons with NaN colors; saving as png only leads to black areas!
        saveFig('test_stereo.svg', drawStereographic(m, rasterized=False))
    
    def _testBasemapBug(self):
        """
        Reproduce the following error:
        File "../mpl_toolkits/basemap/__init__.py", line 1079, in __init__ self._readboundarydata('gshhs',as_polygons=True) 
        File "../mpl_toolkits/basemap/__init__.py", line 1422, in _readboundarydata if not poly.is_valid(): poly=poly.fix() 
        File "_geoslib.pyx", line 234, in _geoslib.BaseGeometry.fix (src/_geoslib.c:2043) 
        File "_geoslib.pyx", line 334, in _geoslib.Polygon.__init__ (src/_geoslib.c:3233) 
        IndexError: index -1 is out of bounds for axis 0 with size 0
        
        see https://github.com/matplotlib/basemap/issues/161
        """
        # it seems that the error only occurs for certain combinations of values,
        # e.g. lon_0=-2 works
        Basemap(width=1000000, height=1000000, projection='stere', lat_0=-88, lon_0=-3)

    def _testMplCrashSimple(self):
        """
        reproduces https://github.com/matplotlib/matplotlib/issues/3304
        Note that this test only fails when an exception is added in mpl's
        _backend_agg.cpp, see https://github.com/matplotlib/matplotlib/issues/3304#issuecomment-50455754
        
        When putting an object outside of the plot limits while using svg and rasterized=True, 
        then apparently matplotlib tries to create a zero size rasterized image which leads to an
        uninitialized pointer being assigned as the resulting svg string buffer.
        """
        fig, ax = plt.subplots()
        circle1=plt.Circle((-10,10),rasterized=True)
        ax.add_artist(circle1)
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        fig.savefig('test.svg')
        plt.close(fig)
       
    
    def _testMiracle(self):
        m = _getMiracleMapping()
        saveFig('heat_intersection.png', drawHeatmap(m.intersectionInflatedCorner[...,0]))
        drawHeatmaps(m)
        m.checkGuarantees()
        m = m.maskedByElevation(10)
        m.checkGuarantees()
        m = resample(m, arcsecPerPx=100, method='mean')
        m.checkGuarantees()
        saveFig('test_stereo.png', drawStereographic(m))

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

def _getMappingNorth():
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMappingSouth():
    imagePath = getResourcePath('ISS029-E-8492.jpg')
    wcsPath = getResourcePath('ISS029-E-8492.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMiracleMapping():
    imagePath = getResourcePath('SOD120304_171900_557_1000.jpg')
    mapping = miracle.getMapping(imagePath)
    return mapping


def _testCoords(offset):
    n = 10
    sp, step = np.linspace(offset,offset+10, num=n, retstep=True)
    coord = np.tile(sp, n).reshape(n,n).astype(np.float32)

    r = n*0.4
    y,x = np.ogrid[-r: r+1, -r: r+1]
    mask = x**2+y**2 <= r**2

    coord[mask] = -coord[mask]
    coord[coord>0] = np.nan
    coord[mask] = -coord[mask]
    
    coordCenter = coord[:-1,:-1] + step/2
    
    return coord, coordCenter
    
def testCoords():
    # no discontinuity or pole
    lats, latsCenter = _testCoords(70)
    lats = lats.T
    latsCenter = latsCenter.T
    
    lons, lonsCenter = _testCoords(160)
    
    return lats, lons, latsCenter, lonsCenter
from astropy.coordinates import Angle
import astropy.units as u
def testCoordsDiscontinuity():
    lats, latsCenter = _testCoords(70)
    lats = lats.T
    latsCenter = latsCenter.T
    
    lons, lonsCenter = _testCoords(160)
    
    lons = Angle((lons + 15) * u.deg).wrap_at(180 * u.deg).degree
    lonsCenter = Angle((lonsCenter + 15) * u.deg).wrap_at(180 * u.deg).degree
    return lats, lons, latsCenter, lonsCenter


# Copyright European Space Agency, 2013

from __future__ import absolute_import, print_function

import numpy as np
from numpy.core.umath_tests import matrix_multiply
from astropy.wcs.wcs import WCS
from auromat.utils import vectorLengths
from auromat.coordinates.transformations import euler_matrix
from auromat.coordinates.transform import cartesian_to_spherical,\
    spherical_to_cartesian
    
try:
    from numexpr import evaluate as ne
except ImportError:
    ne = None

def pix2world(wcsHeader, width, height, startX=0, startY=0, corner=True, ascartesian=False):
    """
    Calculate RA, Dec coordinates of a given pixel coordinate rectangle.
           
    ra, dec = pix2world(..)
    ra and dec are indexed as [y,x]
    
    Each array element contains the RA,Dec coords of the top left corner of the
    given pixel if corner==True, otherwise the coords of the pixel center. 
    If corner==True, an additional row and column exists at the bottom and right so that
    it is possible to get the bottom and right corner values for those pixels.
    
    :param dictionary wcsHeader: WCS header
    :param width: width of rectangle
    :param height: height of rectangle
    :param startX: x coordinate of rectangle, can be negative
    :param startY: y coordinate of rectangle, can be negative
    
    If ascartesian=False:
    
    :rtype: tuple(ra, dec) with arrays of shape (height+1,width+1) if corner==True, else (height,width)
    
    If ascartesian=True:
    
    :rtype: array of shape (height[+1],width[+1],3) with x,y,z order 
    """
    if corner:
        startX -= 0.5 # top left corner instead of pixel center
        startY -= 0.5
    x, y = np.meshgrid(np.arange(startX,startX+width+corner), np.arange(startY,startY+height+corner))
    
    # check if TAN projection and use our fast version, otherwise fall-back to astropy
    if wcsHeader['CTYPE1'] == 'RA---TAN' and wcsHeader['CTYPE2'] == 'DEC--TAN' and \
       wcsHeader['LATPOLE'] == 0.0:
        res = tan_pix2world(wcsHeader, x, y, 0, ascartesian=ascartesian)
        
    else:
        wcs = WCS(wcsHeader)
        ra, dec = wcs.all_pix2world(x, y, 0, ra_dec_order=True)
        if ascartesian:
            np.deg2rad(ra, ra)
            np.deg2rad(dec, dec)
            res = spherical_to_cartesian(None, dec, ra, astuple=False)
        else:
            res = ra, dec
    
    return res

def tan_pix2world(header, px, py, origin, ascartesian=False):
    """
    Fast reimplementation of astropy.wcs.wcs.wcs_pix2world with support for
    only the TAN projection. Speedup is about 2x.
    
    :rtype: tuple (ra,dec) in degrees, or cartesian coordinates in one array (h,w,3) 
            if ascartesian=True
    """
    assert origin in [0,1]
    assert px.shape == py.shape
    shape = px.shape
    px = px.ravel()
    py = py.ravel()
    
    assert header['CTYPE1'] == 'RA---TAN'
    assert header['CTYPE2'] == 'DEC--TAN'
    assert header['LATPOLE'] == 0.0
    lonpole = header['LONPOLE']
    
    ra_ref = header['CRVAL1'] # degrees
    dec_ref = header['CRVAL2']
    px_ref = header['CRPIX1']
    py_ref = header['CRPIX2']
    cd = np.array([[header['CD1_1'], header['CD1_2']],
                   [header['CD2_1'], header['CD2_2']]])
                
    # make pixel coordinates relative to reference point and put them in one array
    pxy = np.empty((len(px),2), float)
    pxy[:,0] = px
    pxy[:,0] -= px_ref
    pxy[:,1] = py
    pxy[:,1] -= py_ref
    if origin == 0:
        pxy += 1
    
    # projection plane coordinates
    xy = matrix_multiply(cd, pxy[...,np.newaxis]).reshape(pxy.shape) # [18% of execution time]
    
    del pxy
    
    # native spherical coordinates
    # spherical projection
    if ne:
        x = xy[:,0]
        y = xy[:,1]
        r = ne('sqrt(x*x+y*y)')
    else:
        r = vectorLengths(xy)
    
    if ne:
        lon = ne('arctan2(x, -y)') # [6% of execution time]
    else:
        lon = np.arctan2(xy[:,0], -xy[:,1])
    del xy
        
    #lat = np.arctan(180/(np.pi*r))
    
    # optimized:
    with np.errstate(divide='ignore'):
        np.reciprocal(r, r)
    np.multiply(180/np.pi, r, r)
    if ne:
        ne('arctan(r)', out=r)
    else:
        np.arctan(r, r)
    lat = r
    
    # celestial spherical coordinates
    # spherical rotation
    euler_z = ra_ref+90
    euler_x = 90-dec_ref
    #euler_z2 = lonpole-90
    euler_z2 = -(lonpole-90) # for some reason, this needs to be negative, contrary to paper
    rotmat = euler_matrix(np.deg2rad(euler_z), np.deg2rad(euler_x), np.deg2rad(euler_z2), 'rzxz')[:3,:3]
    
    lmn = spherical_to_cartesian(None, lat, lon, astuple=False) # [12% of execution time]
    lmnrot = matrix_multiply(rotmat, lmn[...,np.newaxis]).reshape(lmn.shape) # [17% of execution time]
    if ascartesian:
        return lmnrot.reshape(shape + (3,))
    
    dec, ra = cartesian_to_spherical(lmnrot[:,0], lmnrot[:,1], lmnrot[:,2], with_radius=False) # [15% of execution time]
    
    np.rad2deg(dec, dec)    
    np.rad2deg(ra, ra)
    # wrap at 360deg so that values are in [0,360]
    ra -= 360
    np.mod(ra, 360, ra) # [12% of execution time]
    
    ra = ra.reshape(shape)
    dec = dec.reshape(shape)
    
    return ra, dec



# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

from six.moves import map
from datetime import datetime
import os
from collections import OrderedDict
import json
import warnings
import glob
import numpy as np

from auromat.mapping.mapping import BaseMappingProvider
import auromat.utils
from auromat.util.url import downloadFile
from auromat.mapping.spacecraft import getMapping, getMappingSequence
from auromat.util.os import makedirs
from auromat.util.lensdistortion import getLensfunModifierFromParams
from auromat.util.image import saveImage, croppedImage
from auromat.util.decorators import inherit_docs

try:
    import rawpy
    import rawpy.enhance
    import lensfunpy
except ImportError as e:
    print(repr(e))
    warnings.warn('rawpy or lensfunpy is missing, no RAW support available')

defaultBaseUrl = 'http://arrrgh-tools.cosmos.esa.int/api/georef_seqs/'

isoDateFormat = '%Y-%m-%dT%H:%M:%S.%f'

def _parseDates(dic):
    keys = {'date', 'date_start', 'date_end'} & set(dic.keys())
    for k in keys:
        dic[k] = datetime.strptime(dic[k], isoDateFormat)
    return dic

@inherit_docs
class ISSMappingProvider(BaseMappingProvider):
    """
    Provider for ESA's ISS Auroral Photography Mappings.
    """
    def __init__(self, cacheFolder, id_=None, useRaw=True, altitude=110, 
                 sequenceInParallel=False, fastCenterCalculation=False, maxTimeOffset=3,
                 raw_white_balance=None, raw_gamma=(1,1), raw_bps=16, raw_auto_bright=False,
                 noRawPostprocessCaching=True,
                 baseUrl=defaultBaseUrl, offline=False):
        """
        
        :param cacheFolder: folder where images and WCS files are downloaded to
                            Note that each sequence must be in its own folder!
        :param int id_: sequence id, can be omitted in later calls due to caching
        :param useRaw: 
            If True, download raw images and apply necessary pre-processing
            locally if necessary (rotation, lens distortion correction,
            bad pixel removal).
            This requires rawpy and lensfunpy.
            If the sequence is not available in RAW format, then JPEGs will
            be downloaded instead.
        :param altitude: in km
        :param raw_white_balance: (r,g,b) tuple of multipliers for each color.
            If not given, uses white balance from data set (corresponds to daylight). 
        :param raw_gamma: (inv_gamma,toe_slope) tuple. 
            For visually pleasing images, use (1/2.222,4.5), see recommendation BT.709.
            For linear images (photon count corresponds linearly to color values), use (1,1).
        :param raw_bps: 8 or 16, bits per color sample
        :param raw_auto_bright: 
            If True, automatically brightens the image such that 1% of all pixels are
            fully saturated. Note that this may destroy useful image information.
        :param noRawPostprocessCaching: 
            If True, then postprocessed RAW files are not written to disk as .tiff files.
            This saves disk space but requires re-computation if a mapping is requested
            multiple times. If False, then postprocessed images are cached and must be
            deleted with removePostProcessedImages() if different RAW postprocessing
            settings should be used.
        :param baseUrl: API base url to the mapping sequences
        :param offline: if True, then missing data is not automatically downloaded,
                        instead an exception is raised
        """
        if raw_bps == 16 and not noRawPostprocessCaching:
            noRawPostprocessCaching = True
            print('noRawPostprocessCaching=False can currently not be used together with raw_bps=16,'
                  'the parameter is implicitly set to True')
            # This is because there is data corruption when saving 16bit Tiff images
            # with scikit-image (0.11dev) and Pillow (2.6). The exact cause is not known.
            
        BaseMappingProvider.__init__(self, maxTimeOffset=maxTimeOffset)
        makedirs(cacheFolder)
        self.cacheFolder = cacheFolder
        self.noRawPostprocessCaching = noRawPostprocessCaching
        self.offline = offline
        
        self.apiDataPath = os.path.join(cacheFolder, 'api.json')
        if not os.path.exists(self.apiDataPath) and not offline:
            if not id_:
                raise ValueError('The id_ parameter must be given the first time')
            url = baseUrl + str(id_)
            downloadFile(url, self.apiDataPath)
        with open(self.apiDataPath, 'r') as fp:
            self.apiData = json.load(fp, object_hook=_parseDates)
                    
        self.metadataPath = os.path.join(cacheFolder, 'metadata.json')
        if not os.path.exists(self.metadataPath) and not offline:
            downloadFile(self.apiData['metadata_uri'], self.metadataPath)
        with open(self.metadataPath, 'r') as fp:
            self.metadata = json.load(fp, object_hook=_parseDates)
            
        self.apiImages = OrderedDict(sorted(self.apiData['images'].items(), key=lambda k_v: k_v[1]['date']))
        self.useRaw = useRaw and 'raw_extension' in self.apiData
        self.altitude = altitude
        self.sequenceInParallel = sequenceInParallel
        self.fastCenterCalculation = fastCenterCalculation
        
        self.processedImagePaths = {}
        self.wcsPaths = {}
        if self.useRaw:
            self.raw_white_balance = raw_white_balance
            self.raw_gamma = raw_gamma
            self.raw_bps = raw_bps
            self.raw_no_auto_bright = not raw_auto_bright
            self.rawImagePaths = {}
            self.badPixelsPath = os.path.join(cacheFolder, 'bad_pixels.gz')
            if not os.path.exists(self.badPixelsPath) and not offline:
                downloadFile(self.apiData['raw_bad_pixels_uri'], self.badPixelsPath)
            self.badPixels = np.loadtxt(self.badPixelsPath, int)
    
    @property
    def range(self):
        return self.apiData['date_start'], self.apiData['date_end']
        
    def _getIdxWithOffset(self, date):
        dates = [i['date'] for i in self.apiImages.values()]
        
        idx = auromat.utils.findNearest(dates, date)
        offset = abs(dates[idx]-date).total_seconds()
        return idx, offset
        
    def contains(self, date):
        _, offset = self._getIdxWithOffset(date)
        return offset <= self.maxTimeOffset
    
    def get(self, date):
        idx, offset = self._getIdxWithOffset(date)
        if offset > self.maxTimeOffset:
            raise ValueError('No image found')
        
        key = list(self.apiImages.keys())[idx]
        
        self._downloadFiles(key)
        wcsPath = self.wcsPaths[key]
        imagePathOrArray = self._processedImage(key)
        
        metadata = dict(list(self.metadata['sequence_metadata'].items()) +
                        list(self.metadata['image_metadata'][key].items()))
        
        mapping = getMapping(imagePathOrArray, wcsPath, altitude=self.altitude, 
                             fastCenterCalculation=self.fastCenterCalculation,
                             metadata=metadata, identifier=key)
        return mapping
    
    def getById(self, identifier):
        return self.get(self.apiImages[identifier]['date'])
    
    def getSequence(self, dateBegin=None, dateEnd=None):
        keys = self.download(dateBegin, dateEnd)
        
        wcsPaths = [self.wcsPaths[k] for k in keys]
        imagePathsOrArrays = map(self._processedImage, keys)
        
        metadatas = [dict(list(self.metadata['sequence_metadata'].items()) + 
                          list(self.metadata['image_metadata'][k].items()))
                     for k in keys]
        
        return getMappingSequence(imagePathsOrArrays, wcsPaths, altitude=self.altitude, 
                                  parallel=self.sequenceInParallel, 
                                  fastCenterCalculation=self.fastCenterCalculation,
                                  metadatas=metadatas)
    
    def download(self, dateBegin=None, dateEnd=None):
        """
        Download part of or the whole sequence to the cache folder.
        An error is raised if self.offline is True and data has not
        been downloaded yet.
        """
        if not dateBegin:
            dateBegin = self.range[0]
        if not dateEnd:
            dateEnd = self.range[1]
        keys = list(k for k,_ in filter(lambda k_v: dateBegin <= k_v[1]['date'] <= dateEnd, self.apiImages.items()))    
        for key in keys:
            self._downloadFiles(key)
        return keys
    
    def removePostProcessedImages(self):
        """
        Removes postprocessed RAW images.
        Useful when noRawPostprocessCaching=False and different RAW
        postprocessing settings need to be applied.
        """
        if not self.useRaw:
            raise ValueError('useRaw must be True')
        for p in glob.glob(os.path.join(self.cacheFolder, '*.tiff')):
            os.remove(p)
        self.processedImagePaths = {}
    
    def _processedImage(self, key):
        if key in self.processedImagePaths:
            return self.processedImagePaths[key]
        assert self.useRaw
        rawPath = self.rawImagePaths[key]
       
        raw = rawpy.imread(rawPath)           
        rawpy.enhance.repair_bad_pixels(raw, self.badPixels)
        if self.raw_white_balance:
            if not (raw.color_desc == 'RGBG' and raw.num_colors == 3):
                raise NotImplementedError
            wb = [self.raw_white_balance[0], self.raw_white_balance[1], 
                  self.raw_white_balance[2], self.raw_white_balance[1]]
        else:
            wb = self.apiData['raw_white_balance']
        rgb = raw.postprocess(user_wb=wb, output_bps=self.raw_bps, 
                              no_auto_bright=self.raw_no_auto_bright,
                              gamma=self.raw_gamma, user_flip=False)
        
        if self.apiData['raw_is_upside_down']:
            # rotate 180deg
            rgb = rgb[::-1,::-1]
        
        # correct lens distortion
        dist_corr = self.apiData['distortion_correction']
        if dist_corr:
            # TODO undistCoords could be cached to avoid recomputing them each time
            mod = getLensfunModifierFromParams(dist_corr['model'], dist_corr['params'], 
                                               rgb.shape[1], rgb.shape[0])
            undistCoords = mod.apply_geometry_distortion()
            rgb = lensfunpy.util.remap(rgb, undistCoords)
        
        # WCS solutions from the ISS dataset are based on cropped images,
        # therefore we crop here as well to make it match.
        rgb = croppedImage(rgb, divisible_by=16)
        
        if self.noRawPostprocessCaching:
            return rgb
        else:       
            processedImagePath = os.path.join(self.cacheFolder, key + '.tiff')
            saveImage(processedImagePath, rgb)
            self.processedImagePaths[key] = processedImagePath
            return processedImagePath
        
    def _downloadFiles(self, key):
        apiImage = self.apiImages[key]
        imageUrl = apiImage['raw_uri' if self.useRaw else 'image_uri']
        wcsUrl = apiImage['wcs_uri']
        imageExt = self.apiData['raw_extension' if self.useRaw else 'image_extension']

        imagePath = os.path.join(self.cacheFolder, key + imageExt)
        wcsPath = os.path.join(self.cacheFolder, key + '.wcs')
        
        if not os.path.exists(imagePath):        
            assert not self.offline
            downloadFile(imageUrl, imagePath)
            
        if not os.path.exists(wcsPath):
            assert not self.offline
            downloadFile(wcsUrl, wcsPath)
        
        if self.useRaw:
            self.rawImagePaths[key] = imagePath
        else:
            self.processedImagePaths[key] = imagePath
        self.wcsPaths[key] = wcsPath        
    